"""
Distances
---------

All the functions here are functions to find the distance between two pieces of data
based on contingency table.

+--------+-----------------+
|        | Row 2           |
+--------+-----+-----+-----+
| Row 1  |     | 1   | 0   |
|        +-----+-----+-----+
|        | 1   | a   | b   |
|        +-----+-----+-----+
|        | 0   | c   | d   |
+--------+-----+-----+-----+

Each function has the same parameters and return values:

Parameters
    a : int
        Count of same 1 numbers.
    b : int
        The sum of 1 in row 1 and 0 in row 2.
    c : int
        Sum 1 in row 2 and 0 in row 1.
    d : int
        Count of same 0 numbers.
    n : int
        Sum of all the numbers.

Returns
    float
        The similarity value between two pieces of data.
"""

from math import sqrt
import warnings

# function to get all the name of the functions
def get_all_functions_name():
    """
    Get all the functions in the module.

    Returns
    -------
    list
        List of functions object.
    """
    funcs = [func for func in globals() if callable(globals()[func])]
    # return excluded list from math and this function
    return funcs[2:]

# get specific function by name
def get_function(name):
    """
    Get function by name.

    Parameters
    ----------
    name : str
        Name of the function.

    Returns
    -------
    function
        Function object.
    """
    return globals()[name]

def hamming(a, b, c, d, n):
    """
    Hamming coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = b+c
        return result
    except:
        return None

def euclidean(a, b, c, d, n):
    """
    Euclidean distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = sqrt(b+c)
        return result
    except:
        return None

def squared_euclidean(a, b, c, d, n):
    """
    Squared Euclidean distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = sqrt((b+c)**2)
        return result
    except:
        return None

def canberra(a, b, c, d, n):
    """
    Canberra distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)**(2/3)
        return result
    except:
        return None

def manhattan(a, b, c, d, n):
    """
    Manhattan distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)
        return result
    except:
        return None

def mean_manhattan(a, b, c, d, n):
    """
    Mean Manhattan distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)/(a+b+c+d)
        return result
    except:
        return None

def cityblock(a, b, c, d, n):
    """
    Cityblock distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)
        return result
    except:
        return None

def minkowski(a, b, c, d, n):
    """
    Minkowski distance from contingency table
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)**1
        return result
    except:
        return None

def vari(a, b, c, d, n):
    """
    Vari distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)/(4*(a+b+c+d))
        return result
    except:
        return None

def size_difference(a, b, c, d, n):
    """
    Size difference from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)**2/((a+b+c+d)**2)
        return result
    except:
        return None

def shape_difference(a, b, c, d, n):
    """
    Shape difference from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (n*(b+c)-((b+c)**2))/((a+b+c+d)**2)
        return result
    except:
        return None

def pattern_difference(a, b, c, d, n):
    """
    Pattern difference from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 4*b*c/((a+b+c+d)**2)
        return result
    except:
        return None

def lance_williams(a, b, c, d, n):
    """
    Lance Williams distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)/(2*a+b+c)
        return result
    except:
        return None

def bray_curtis(a, b, c, d, n):
    """
    Bray-Curtis distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (b+c)/(2*a+b+c)
        return result
    except:
        return None

def hellinger(a, b, c, d, n):
    """
    Hellinger distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*sqrt(1-(a/sqrt((a+b)*(a+c))))
        return result
    except:
        return None

def chord(a, b, c, d, n):
    """
    Chord distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = sqrt(2*(1-(a/sqrt((a+b)*(a+c)))))
        return result
    except:
        return None

def yuleq(a, b, c, d, n):
    """
    Yule-Q distance from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')
    
    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (2*b*c)/(a*d + b*c)
        return result
    except:
        return None


"""
Similarity
----------

All the functions here are functions to find similarities between two pieces of data
based on contingency table.

+--------+-----------------+
|        | Row 2           |
+--------+-----+-----+-----+
| Row 1  |     | 1   | 0   |
|        +-----+-----+-----+
|        | 1   | a   | b   |
|        +-----+-----+-----+
|        | 0   | c   | d   |
+--------+-----+-----+-----+

Each function has the same parameters and return values:

Parameters
    a : int
        Count of same 1 numbers.
    b : int
        The sum of 1 in row 1 and 0 in row 2.
    c : int
        Sum 1 in row 2 and 0 in row 1.
    d : int
        Count of same 0 numbers.
    n : int
        Sum of all the numbers.

Returns
    float
        The similarity value between two pieces of data.
"""

from math import log2, sqrt, pi, cos, log10
import numpy as np
import warnings

# function to get all the name of the functions
def get_all_functions_name():
    """
    Get all the functions in the module.

    Returns
    -------
    list
        List of all the functions.
    """
    funcs = [func for func in globals() if callable(globals()[func])]
    # return excluded list from math and this function
    return funcs[6:]

# function to get specific function by name
def get_function(name):
    """
    Get function by name.

    Parameters
    ----------
    name : str
        Name of the function.

    Returns
    -------
    function
        Function object.
    """
    return globals()[name]

def jaccard(a, b, c, d, n):
    """
    Jaccard coefficient from contingency table.
    """
    
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/(a+b+c)
        return result
    except:
        return None

def dice_2(a, b, c, d, n):
    """
    Dice 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/(2*a+b+c)
        return result
    except:
        return None

def dice_1(a, b, c, d, n):
    """
    Dice 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*a/(2*a+b+c)
        return result
    except:
        return None

def three_way_jaccard(a, b, c, d, n):
    """
    Three-way Jaccard coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 3*a/(3*a+b+c)
        return result
    except:
        return None

def neili(a, b, c, d, n):
    """
    Neili coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*a/((a+b)+(a+c))
        return result
    except:
        return None

def sokal_sneath_1(a, b, c, d, n):
    """
    Sokal-Sneath 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/(a+2*b+2*c)
        return result
    except:
        return None

def sokal_michener(a, b, c, d, n):
    """
    Sokal-Michener coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a+d)/(a+b+c+d)
        return result
    except:
        return None

def sokal_sneath_2(a, b, c, d, n):
    """
    Sokal-Sneath 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*(a+d)/(2*a+b+c+2*d)
        return result
    except:
        return None

def rogers_tanimoto(a, b, c, d, n):
    """
    Rogers-Tanimoto coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a+d)/(a+2*(b+c)+d)
        return result
    except:
        return None

def faith(a, b, c, d, n):
    """
    Faith coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a+0.5*d)/(a+b+c+d)
        return result
    except:
        return None

def gower_legendre(a, b, c, d, n):
    """
    Gower-Legendre coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a+d)/(a+0.5*(b+c)+d)
        return result
    except:
        return None

def intersection(a, b, c, d, n):
    """
    Intersection coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a
        return result
    except:
        return None

def inner_product(a, b, c, d, n):
    """
    Inner product coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a+d
        return result
    except:
        return None

def russell_rao(a, b, c, d, n):
    """
    Russell-Rao coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/(a+b+c+d)
        return result
    except:
        return None

def cosine(a, b, c, d, n):
    """
    Cosine coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/sqrt((a+b)*(a+c))
        return result
    except:
        return None

def gilbert_wells(a, b, c, d, n):
    """
    Gilbert-Wells coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = log2(a) - log2(n) - log2((a+b)/n) - log2((a+c)/n)
        return result
    except:
        return None

def ochiai_1(a, b, c, d, n):
    """
    Ochiai 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/sqrt((a+b)*(a+c))
        return result
    except:
        return None

def forbes_1(a, b, c, d, n):
    """
    Forbes 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = n*a/((a+b)*(a+c))
        return result
    except:
        return None

def fossum(a, b, c, d, n):
    """
    Fossum coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = n*((a-0.5)**2)/((a+b)*(a+c))
        return result
    except:
        return None

def sorgenfrei(a, b, c, d, n):
    """
    Sorgenfrei coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a**2/((a+b)*(a+c))
        return result
    except:
        return None

def mountford(a, b, c, d, n):
    """
    Mountford coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/(0.5*(a*b+a*c)+b*c)
        return result
    except:
        return None

def otsuka(a, b, c, d, n):
    """
    Otsuka coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/sqrt((a+b)*(a+c))
        return result
    except:
        return None

def mc_connaughey(a, b, c, d, n):
    """
    McConnaughey coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a**2-b*c)/((a+b)*(a+c))
        return result
    except:
        return None

def tarwid(a, b, c, d, n):
    """
    Tarwid coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (n*a - (a+b)*(a+c))/(n*a + (a+b)*(a+c))
        return result
    except:
        return None

def kulczynski_2(a, b, c, d, n):
    """
    Kulczynski 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a/2 * (2*a+b+c))/((a+b)*(a+c))
        return result
    except:
        return None

def driver_kroeber(a, b, c, d, n):
    """
    Driver-Kroeber coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a/2)*(1/(a+b) + 1/(a+c))
        return result
    except:
        return None

def johnson(a, b, c, d, n):
    """
    Johnson coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/(a+b) + a/(a+c)
        return result
    except:
        return None

def dennis(a, b, c, d, n):
    """
    Dennis coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a*d - b*c)/sqrt(n*(a+b)*(a+c))
        return result
    except:
        return None

def simpson(a, b, c, d, n):
    """
    Simpson coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/min(a+b, a+c)
        return result
    except:
        return None

def braun_banquet(a, b, c, d, n):
    """
    Braun-Banquet coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a/max(a+b, a+c)
        return result
    except:
        return None

def fager_mcgowan(a, b, c, d, n):
    """
    Fager-McGowan coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a/sqrt((a+b)*(a+c))) - max(a+b, a+c)/2
        return result
    except:
        return None

def forbes_2(a, b, c, d, n):
    """
    Forbes 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (n*a - (a+b)*(a+c))/(n*min(a+b, a+c) - (a+b)*(a+c))
        return result
    except:
        return None

def sokal_sneath_4(a, b, c, d, n):
    """
    Sokal-Sneath 4 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a/(a+b) + a/(a+c) + d/(b+d) + d/(c+d))/4
        return result
    except:
        return None

def gower(a, b, c, d, n):
    """
    Gower coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a+d)/sqrt((a+b)*(a+c)*(b+d)*(c+d))
        return result
    except:
        return None

def pearson_1(a, b, c, d, n):
    """
    Pearson 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (n*(a*d - b*c)**2)/((a+b)*(a+c)*(b+d)*(c+d))
        return result
    except:
        return None

def pearson_2(a, b, c, d, n):
    """
    Pearson 2 coefficient from contingency table.
    """
    x2 = pearson_1(a, b, c, d, n)
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = sqrt(x2/(n + x2))
        return result
    except:
        return None

def pearson_3(a, b, c, d, n):
    """
    Pearson 3 coefficient from contingency table.
    """
    rho = (a*d - b*c)/sqrt((a+b)*(a+c)*(b+d)*(c+d))
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = sqrt(rho/(n + rho))
        return result
    except:
        return None

def pearson_heron_1(a, b, c, d, n):
    """
    Pearson Heron 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a*d - b*c)/sqrt((a+b)*(a+c)*(b+d)*(c+d))
        return result
    except:
        return None

def pearson_heron_2(a, b, c, d, n):
    """
    Pearson Heron 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = cos((pi * sqrt(b*c)/(sqrt(a*d) + sqrt(b*c))))
        return result
    except:
        return None

def sokal_sneath_3(a, b, c, d, n):
    """
    Sokal-Sneath 3 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a+d)/(b+c)
        return result
    except:
        return None

def sokal_sneath_5(a, b, c, d, n):
    """
    Sokal-Sneath 5 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a*d/((a+b)*(a+c)*(b+d)*sqrt(c+d))
        return result
    except:
        return None

def cole(a, b, c, d, n):
    """
    Cole coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = sqrt(2)*(a*d - b*c)/sqrt((a*d - b*c)**2 - (a+b)*(a+c)*(b+d)*(c+d))
        return result
    except:
        return None

def stiles(a, b, c, d, n):
    """
    Stiles coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = log10(n*(abs(a*d - b*c) - n/2)**2 / ((a+b)*(a+c)*(b+d)*(c+d)))
        return result
    except:
        return None

def ochiai_2(a, b, c, d, n):
    """
    Ochiai 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a*d/sqrt((a+b)*(a+c)*(b+d)*(c+d))
        return result
    except:
        return None

def yuleq(a, b, c, d, n):
    """
    Yuleq coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a*d - b*c)/(a*d + b*c)
        return result
    except:
        return None

def yulew(a, b, c, d, n):
    """
    Yulew coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (sqrt(a*d) - sqrt(b*c))/(sqrt(a*d) + sqrt(b*c))
        return result
    except:
        return None

def kulczynski_1(a, b, c, d, n):
    """
    Kulczynski 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a / (b + c)
        return result
    except:
        return None

def tanimoto(a, b, c, d, n):
    """
    Tanimoto coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a / ((a+b) + (a+c) - a)
        return result
    except:
        return None

def disperson(a, b, c, d, n):
    """
    Dispersion coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a*d - b*c)/(a+b+c+d)**2
        return result
    except:
        return None

def hamann(a, b, c, d, n):
    """
    Hamann coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = ((a+d) - (b+c))/(a+b+c+d)
        return result
    except:
        return None

def michael(a, b, c, d, n):
    """
    Michael coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 4*(a*d - b*c)/((a+d)**2 + (b+c)**2)
        return result
    except:
        return None

def goodman_kruskal(a, b, c, d, n):
    """
    Goodman-Kruskal coefficient from contingency table.
    """
    tho = max(a,b) + max(c,d) + max(a,c) + max(b,d)
    tho_diff = max(a+c, b+d) + max(a+b, c+d)
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (tho - tho_diff)/(2*n - tho_diff)
        return result
    except:
        return None

def anderberg(a, b, c, d, n):
    """
    Andersberg coefficient from contingency table.
    """
    tho = max(a,b) + max(c,d) + max(a,c) + max(b,d)
    tho_diff = max(a+c, b+d) + max(a+b, c+d)
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (tho - tho_diff)/(2*n)
        return result
    except:
        return None

def baroni_urbani_buser_1(a, b, c, d, n):
    """
    Baroni-UrbaniBuser 1 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (sqrt(a*d) + a)/(sqrt(a*d) + a + b + c)
        return result
    except:
        return None

def baroni_urbani_buser_2(a, b, c, d, n):
    """
    Baroni-UrbaniBuser 2 coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (sqrt(a*d) + a - (b+c))/(sqrt(a*d) + a + b + c)
        return result
    except:
        return None

def peirce(a, b, c, d, n):
    """
    Peirce coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a*b + b*c)/(a*b + 2*b*c + c*d)
        return result
    except:
        return None

def eyraud(a, b, c, d, n):
    """
    Eyraud coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = n**2 * (n*a - (a+b)*(a+c))/((a+b)*(a+c)*(b+d)*(c+d))
        return result
    except:
        return None

def tarantula(a, b, c, d, n):
    """
    Tarantula coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a*(c+d)/(c*(a+b))
        return result
    except:
        return None

def ample(a, b, c, d, n):
    """
    Ample coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = abs(tarantula(a, b, c, d, n))
        return result
    except:
        return None

def derived_rusell_rao(a, b, c, d, n):
    """
    Derived Rusell-Rao coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = log2(1+a)/log2(1+n)
        return result
    except:
        return None

def derived_jaccard(a, b, c, d, n):
    """
    Derived Jaccard coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = log2(1+a)/log2(1+a+b+c)
        return result
    except:
        return None

def var_of_correlation(a, b, c, d, n):
    """
    Variance of correlation coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (log2(1+a*d) - log2(1+b*c))/log2(1+n**2/4)
        return result
    except:
        return None

def gleason(a, b, c, d, n):
    """
    Gleason coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*a / (2*a+b+c)
        return result
    except:
        return None

def van_der_maarel(a, b, c, d, n):
    """
    Van der Maarel coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (2*a-b-c)/(2*a+b+c)
        return result
    except:
        return None

def consonni_and_todeschini_iv(a, b, c, d, n):
    """
    Consonni and Todeschini (CT IV) coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = np.log(1+a)/np.log(1+a+b+c)
        return result
    except:
        return None

def consonni_and_todeschini_iii(a, b, c, d, n):
    """
    Consonni and Todeschini (CT III) coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = np.log(1+a)/np.log(1+n)
        return result
    except:
        return None

def austin_and_colwell(a, b, c, d, n):
    """
    Austin and Colwell coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (2/np.pi) * (1/np.sin(sqrt((a+d)/n)))
        return result
    except:
        return None

def consonni_and_todeschini_i(a, b, c, d, n):
    """
    Consonni and Todeschini (CT I) coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = np.log(1+a+d)/np.log(1+n)
        return result
    except:
        return None

def phi(a, b, c, d, n):
    """
    Phi coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a*d - b*c)/sqrt((a+b)*(b+d)*(a+c)*(c+d))
        return result
    except:
        return None

def cohen(a, b, c, d, n):
    """
    cohen coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*(a*d - b*c) / sqrt((a+b)*(b+d)+(a+c)*(c+d))
        return result
    except:
        return None

def maxwell_and_pilliner(a, b, c, d, n):
    """
    Maxwell and Pilliner coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = 2*(a*d - b*c) / sqrt((a+b)*(c+d)+(a+c)*(b+d))
        return result
    except:
        return None

def consonni_and_todeschini_v(a, b, c, d, n):
    """
    Consonni and Todeschini (CT V) coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (np.log(1+a*d) - np.log(1+b*c))/np.log(1+(n**2)/4)
        return result
    except:
        return None

def scott(a, b, c, d, n):
    """
    scott coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (4*a*d - (b+c)**2) / ((2*a+b+c)*(2*d+b+c))
        return result
    except:
        return None

def tetrachoric(a, b, c, d, n):
    """
    Tetrachoric coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = np.cos((180)/(1+sqrt((a*d)/(b*c))))
        return result
    except:
        return None

def odds_ratio(a, b, c, d, n):
    """
    odds ratio coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = a*d / (b*c)
        return result
    except:
        return None

def rand(a, b, c, d, n):
    """
    rand coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        N = n*(n-1)/2
        B = a*b + c*d
        C = a*c + b*d
        D = a*d + b*c
        A = N-B-C-D
        result = (A+B)/N
        return result
    except:
        return None

def ARI(a, b, c, d, n):
    """
    ARI coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        N = n*(n-1)/2
        B = a*b + c*d
        C = a*c + b*d
        D = a*d + b*c
        A = N-B-C-D
        result = (abs(N*(A+D) - abs((A+B)*(A+C) + (C+D)*(B+D))))/abs(N**2 - abs((A+B)*(A+C) + (C+D)*(B+D)))
        return result
    except:
        return None

def loevingers_H(a, b, c, d, n):
    """
    loevinger's H coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        p1 = max(a, b) + max(c, d) + max(a, c) + max(b, d)
        p2 = max(a+c, b+d) + max(a+b, c+d)
        result = 1 - (b/(n*p1*p2))
        return result
    except:
        return None

def rogot_and_goldberg(a, b, c, d, n):
    """
    rogot and goldberg coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a/(2*a+b+c)) + d/(2*d+b+c)
        return result
    except:
        return None

def hawkins_and_dotson(a, b, c, d, n):
    """
    hawkins and dotson coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (a/(a+b+c) + d/(b+c+d))/2
        return result
    except:
        return None

def harris_and_lahey(a, b, c, d, n):
    """
    harris and lahey coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (2*(2*d+b+c)/(2*(a+b+c)) + d*(2*a+b+c)/(2*(b+c+d)))
        return result
    except:
        return None

def consonni_and_todeschini_ii(a, b, c, d, n):
    """
    Consonni and Todeschini (CT II) coefficient from contingency table.
    """
    if a < 0 or b < 0 or c < 0 or d < 0 or n < 0:
        raise ValueError('value in confusion matrix cannot less than zero')

    # raise warning if a, b, c, d or n is not integer
    if not isinstance(a, int) or not isinstance(b, int) or not isinstance(c, int) or not isinstance(d, int) or not isinstance(n, int):
        warnings.warn(RuntimeWarning('some value in confusion matrix is not integer'))

    try:
        result = (np.log(1+n) - np.log(1+b+c))/np.log(1+n)
        return result
    except:
        return None


"""
Utility
-------

module to store the helper functions used in the main class.
"""

import numpy as np
import pandas as pd

# function to convert similarity value to distance value
def convertEquation(val, type = 1, gamma = 1, eq_type='S'):
    """Convert Equation Value.

    A function to convert similarity value to distance value or vice versa.

    Parameters
    ----------  
    val : float
        similarity value

    type : int, optional
        1 for using squared basic conversion, 2 for using basic convertion, 3 for using gamma

    gamma : float, optional
        gamma value for gamma conversion

    eq_type : str, optional
        initial type of val. 'S' for similarity, 'D' for distance

    Returns
    -------
    float
        distance value

    Raises
    ------
    ValueError
        If type is not 1, 2, or 3
        if eq_type is not 'S' or 'D'

    Examples
    --------
    >>> convertEquation(0.5, 3, 1, 'S')
    0.6065306597126334
    """
    # check val either float or int
    # if not isinstance(val, (float, int)):
    #     raise ValueError('val must be float or int')

    # check gamma, gamma should between 0 and 1
    if gamma < 0 or gamma > 1:
        raise ValueError('Gamma should between 0 and 1')

    # check type
    if type not in [1, 2, 3]:
        raise ValueError('Type should be 1, 2, or 3')

    if type == 1:
        return 1 - val
    elif type == 2:
        return (1 - val) ** 2
    if eq_type.lower() == 'd':
        if type == 3:
            return np.log(val) / (-gamma)
    elif eq_type.lower() == 's':
        if type == 3:
            return np.exp(-val * gamma)
    
    return Exception('eq_type should be "S" for similarity or "D" for distance')

def minMaxNormalization(df):
    """Min Max Normalization.

    Normalize every column in dataframe with min max normalization

    Parameters
    ----------
    df : pandas.DataFrame, list, or pandas.Series
        dataframe to be normalized

    Returns
    -------
    pandas.DataFrame (np.array if list or pandas.Series)
        normalized dataframe

    Examples
    --------
    >>> df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})

    >>> df_normalize = minMaxNormalization(df)

    >>> df_normalize
        A    B
    0  0.0  0.0
    1  0.5  0.5
    2  1.0  1.0
    """
    # if input is list
    if isinstance(df, list) or isinstance(df, pd.Series):
        _array = np.array(df)
        output = (_array - min(_array))/(max(_array) - min(_array))

        return output

    # check df
    if not isinstance(df, pd.DataFrame):
        raise ValueError('df must be pandas.DataFrame or list')

    df_min_max_scaled = df.copy()
    for column in df_min_max_scaled.columns:
        df_min_max_scaled[column] = (df_min_max_scaled[column] - df_min_max_scaled[column].min()) / (df_min_max_scaled[column].max() - df_min_max_scaled[column].min())

    return df_min_max_scaled

def random_sampling_data(df_source, sample_rate = 0.1, seed = None):
        """
        Sample dataframe to get the best result.
        
        Parameters
        ----------
        df_source : pandas.DataFrame
            Dataframe source
        sample_size : float, optional
            The sample rate. The default is 0.1.
        
        Returns
        -------
        df_sample : pandas.DataFrame
            Dataframe that contains the sample dataframe
        """
        # initiate return dataframe
        df_sample = pd.DataFrame()

        # check df
        if not isinstance(df_source, pd.DataFrame):
            raise Exception('df_source must be pandas.DataFrame')

        if sample_rate <= 0 or sample_rate > 1:
                raise Exception('sample_rate must between 0 and 1')

        # get the sample dataframe
        # set seed
        if isinstance(seed, int):
            df_sample = df_source.sample(frac = sample_rate, random_state=seed)
        elif seed == None:
            df_sample = df_source.sample(frac = sample_rate)
        else:
            raise Exception('Seed must none or int')

        return df_sample

def stratified_sampling_data(df_source, sample_rate = 0.1, seed=None):
    """
    Stratified Sampling from data. Make sure your target data is on last column of your dataframe.

    Parameters
    ----------
    df_source: pandas.DataFrame
        Dataframe source
    sample_rate: float
        rate of sample in df_source.
    seed: int, optional
        Seed used in sampling. Default is None.
    
    Returns
    -------
    df_sample: pandas.DataFrame
        Result of sampling in DataFrame form
    """
    df_sample = pd.DataFrame()

    # check df
    if not isinstance(df_source, pd.DataFrame):
        raise Exception('df_source must be pandas.DataFrame')

    if sample_rate <= 0 or sample_rate > 1:
        raise Exception('sample_rate must between 0 and 1')

    n_sample = int(len(df_source) * sample_rate)
    target = df_source.iloc[:, -1].unique()
    n_sample_strata = n_sample//len(target)

    cluster_df = []
    for t in target:
        _df_cluster = df_source[df_source.iloc[:, -1] == t]
        # set seed
        if isinstance(seed, int):
            cluster_df.append(_df_cluster.sample(n = n_sample_strata, random_state=seed))
        elif seed == None:
            cluster_df.append(_df_cluster.sample(n = n_sample_strata))
        else:
            raise Exception('Seed must none or int')

    df_sample = pd.concat(cluster_df)

    return df_sample


"""
Pair Best Similarity/Distance.
---------------------------------------

This module contains classes to find the best similarity/distance equation based on the target match.
"""

import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix, roc_auc_score
from tqdm.autonotebook import tqdm
import random
import measurenary.similarity as sim
import measurenary.distance as dis
from measurenary.utility import *

# disable runtime warning
import warnings
warnings.filterwarnings('ignore', category=RuntimeWarning)

class PairBestMeasure():
    """
    Class to get the best similarity and distance based on target pairs.

    Parameters
    ----------
    show_result : bool, optional
        Set True to show result. The default is False.
    result_count : int, optional
        Set the number of result to show. The default is 5.
    """
    def __init__(self, show_result: bool = False, result_count: int = 5):
        # set up all similarity and dissimilarity equation
        self.f_sim = [val for _, val in sim.__dict__.items() if callable(val)][6:]
        self.f_dis = [val for _, val in dis.__dict__.items() if callable(val)][3:]
        self.name_f_sim = [eq.__name__ + ' similarity' for eq in self.f_sim]
        self.name_f_dis = [eq.__name__ + ' distance' for eq in self.f_dis]
        self.result_count = result_count
        self.show_result = show_result

    def _find_confusion_matrix(self, df_source: pd.DataFrame, index_start: int, index_end: int) -> np.ndarray:
        """
        method to generate a confusion matrix and similarity/distance results from every record pair in data binary dataframe on specific index.

        Parameters
        ----------
        df_source : pandas.DataFrame
            dataframe source
        index_start : int
            start index
        index_end : int
            end index

        Returns
        -------
        list
            List of confusion matrix and similarity/distance values.
        """
        # # initiate return list
        row_total = index_end//2 * (1 + index_end)

        # loop for every row in data
        for index, row in tqdm(df_source.iloc[index_start:, :].iterrows()):
            if index == index_end:
                break
            # print('sample', index, 'of', index_end)
            for _index, _row in tqdm(df_source.iloc[index+1:, :].iterrows(), leave=False):
                # generate a confusion matrix
                cm = confusion_matrix(row[:-1].to_list(), _row[:-1].to_list()).ravel()
                # match the reference column
                ## 1 if match
                ## 0 if it doesn't match
                is_match = True if int(df_source.iloc[index, -1]) == int(df_source.iloc[_index, -1]) else False

                # calculate the similarity/distance value based on confusion matrix before
                _sim = []
                for _, f in enumerate(self.f_sim):
                    try:
                        sim = f(cm[3], cm[1], cm[2], cm[0], sum(cm))
                        if sim == None: 
                            sim = np.nan
                    except:
                        sim = np.nan

                    _sim.append(sim)

                for _, f in enumerate(self.f_dis):
                    try:
                        dis = 1 - (f(cm[3], cm[1], cm[2], cm[0], sum(cm)))**2
                        if dis == None:
                            dis = np.nan
                    except:
                        dis = np.nan
                    _sim.append(dis)

                # combine confusion matrix and similarity/distance result
                yield [index, _index, cm.ravel(), is_match] + _sim

    def fit(self, df: pd.DataFrame, use_seed: bool = False, num_sample: int = 20, **kwargs):
        """
        Train data to generate a suitable similarity/distance equation.

        Parameters
        ----------
        df : pandas.DataFrame
            dataframe source
        use_seed : bool, optional
            Set True to use seed. The default is False.
        num_sample : int, optional
            Set the number of sample to generate. The default is 20.

        Returns
        -------
        None
        """
        # set up show result and result count in kwargs
        if 'show_result' in kwargs:
            self.show_result = kwargs['show_result']
        if 'result_count' in kwargs:
            self.result_count = kwargs['result_count']

        # check boolean value of use_seed
        if not isinstance(use_seed, bool):
            raise Exception('use_seed must be boolean')

        # check df is pandas dataframe
        if not isinstance(df, pd.DataFrame):
            raise Exception('df must be pandas dataframe')

        # check num sample
        if not isinstance(num_sample, int):
            raise Exception('num_sample must be integer')
        if num_sample <= 0:
            raise Exception('num_sample must be greater than 0')
        
        # set up index start and end
        index_start = 0
        index_end = df.shape[0]

        res_list = list(self._find_confusion_matrix(df, index_start, index_end))
        res_arr = np.array(res_list, dtype=object)

        # print(res_arr.shape)

        # set up equation function name
        name_f_sim, name_f_dis = self.name_f_sim, self.name_f_dis

        # create a dataframe from _find_confusion_matrix list result
        res_df = pd.DataFrame(res_arr, columns=['i', 'j', 'confussion_matrix', 'is_match'] + name_f_sim + name_f_dis)

        # get value of similarity/distance measurement
        sim_df = pd.DataFrame(res_df.iloc[:, 4:], columns=name_f_sim + name_f_dis)

        # change inf and -inf value to np.nan
        sim_df.replace([np.inf, -np.inf], np.nan, inplace=True)

        # get column name that has nan in sim_df and drop it
        column_nan = sim_df.columns[sim_df.isnull().any()].tolist()
        sim_df.drop(column_nan, axis=1, inplace=True)

        # add column name that has nan
        nan_value_equation = column_nan

        # print excluded equation that produce nan
        if len(nan_value_equation) != 0:
            name_f_sim = [e for e in name_f_sim if e not in nan_value_equation]
            name_f_dis = [e for e in name_f_dis if e not in nan_value_equation]
            # remove column
            res_df.drop(list(nan_value_equation), axis=1, inplace=True)

        sim_df = pd.concat([sim_df, res_df['is_match']], axis=1)

        # separate is_match from scaled_sim_df to true_df and false_df
        true_df = sim_df[sim_df['is_match'] == True]
        false_df = sim_df[sim_df['is_match'] == False]

        # get length of 1's and 0's
        n_true = len(true_df)
        n_false = len(false_df)
        # drop is_match column and transform to list
        auc_list = [sim_df.drop('is_match', axis=1).columns.to_list()]
        seed_list = []
        for i in tqdm(range(num_sample)):
            # set seed
            if use_seed:
                seed = i * 10
                random.seed(seed)
                seed_list.append(seed)

            # handle imbalance data
            if n_true <= n_false:
                sample_index = random.sample(range(0, n_false), n_true)
                # sample_index = list(range(n_true))
                false_sample_df = false_df.iloc[sample_index, :]
                combined_df = pd.concat([true_df, false_sample_df])
            else:
                sample_index = random.sample(range(0, n_true), n_false)
                # sample_index = list(range(n_false))
                true_sample_df = true_df.iloc[sample_index, :]
                combined_df = pd.concat([true_sample_df, false_df])

            _auc_list = []
            for column in combined_df.drop('is_match', axis=1).columns:
                # compute min max score
                score = minMaxNormalization(combined_df[column])

                roc_auc = roc_auc_score(combined_df['is_match'].astype(int), score)
                _auc_list.append(roc_auc)
            
            auc_list.append(_auc_list)
        
        auc_arr = np.array(auc_list)
        auc_df = pd.DataFrame(auc_arr.T, columns=['sim/dis name'] + ['iter %d' % (i+1) for i in range(num_sample)])
        auc_df['mean_auc']  = auc_df.iloc[:, 1:].astype(float).mean(axis=1)
        auc_df = auc_df.sort_values(by='mean_auc', ascending=False, ignore_index=True)

        self.res_df = res_df
        self.sim_df = sim_df
        self.auc_df = auc_df
        self.seed_list = seed_list
        self.nan_value_equation = nan_value_equation

        # print result if self.show_result is True
        if self.show_result:
            print('\nfinal {} best similarity: '.format(str(self.result_count)))
            print(auc_df[['sim/dis name', 'mean_auc']].head(self.result_count))

    def get_result(self, csv: bool = False) -> pd.DataFrame:
        """
        Get result of training.

        Parameters
        ----------
        csv : bool, optional
            Set True to get result in csv format. The default is False.
        
        Returns
        -------
        pandas.DataFrame
            dataframe result
        
        Raises
        ------
        Exception
            If method fit() is not called before this function.
        """

        if self.auc_df.empty:
            raise Exception('Fit your data first with \'fit\' method')
        if csv:
            self.auc_df.to_csv('result.csv', index=False)
        
        return self.auc_df




"""
Agglomerative Best Similarity/Distance.
---------------------------------------

This module contains classes to find the best similarity/distance equation based on Agglomerative Clustering.
"""

import measurenary.similarity as sim
import measurenary.distance as dis
import measurenary.utility as util
import pandas as pd
import numpy as np
from sklearn.metrics import confusion_matrix, homogeneity_completeness_v_measure
from sklearn.metrics.cluster import adjusted_rand_score
from scipy.cluster.hierarchy import fclusterdata
from tqdm.autonotebook import tqdm
from typing import Union

# class to get best similarity/distance measurement based on Agglomerative Clustering Algorithm
class AgglomerativeBestMeasure():
    """
    A class to get best usage of similarity/distance with Agglomerative Clustering.

    Parameters
    ----------
    show_result : bool, optional
        True if you want to show the result. The default is False.
    result_count : int, optional
        The number of result to print out. The default is 5.
    """

    def __init__(self, show_result: bool = False, result_count: int = 5):
        # set up all similarity and distance equation
        self.f_sim = [val for _, val in sim.__dict__.items() if callable(val)][6:]
        self.f_dis = [val for _, val in dis.__dict__.items() if callable(val)][3:]
        self.linkage = ['complete', 'average', 'single', 'weighted']
        self.linkage_df = []
        self.show_result = show_result
        self.result_count = result_count

    def fit(self, df: pd.DataFrame, n_clusters = 2, affinity = 'all', linkage = 'all', use_sampling = 'none', sample_rate = 0.1, **kwargs):
        """
        Fit data with Agglomerative Clustering.
        
        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe to fit with Agglomerative Clustering
        n_clusters : int, optional
            Number of cluster to generate, by default 2
        affinity : str, optional
            Type of affinity to use, by default 'all'
        linkage : str, optional
            Type of linkage to use, by default 'all'
        use_sampling : str, optional
            Sampling method that used to reduce computation time by sampling the data. It can be 'none' for not to implement sampling, 'random' to implement random sampling, and 'stratified' for implement stratified sampling
        sample_rate : float, optional
            Sampling rate to determine size of sample. Value from 0 to 1
        
        Returns
        -------
        None
        """
        # set up show result and result count in kwargs
        if 'show_result' in kwargs:
            self.show_result = kwargs['show_result']
        if 'result_count' in kwargs:
            self.result_count = kwargs['result_count']
        # check seed
        if 'seed' in kwargs:
            seed = kwargs['seed']

       # check if df is DataFrame
        if not isinstance(df, pd.DataFrame):
            raise Exception('df must be DataFrame')
        # check if n_clusters is greater than 1
        if n_clusters <= 1:
            raise Exception('n_clusters must be greater than 1')

        # check sample rate
        if sample_rate < 0 or sample_rate > 1:
            raise Exception('sample_rate must be between 0 and 1')

        affinity_sim, affinity_dis = [], []

        # check if affinity is valid
        if isinstance(affinity, str):
            if affinity == 'all':
                affinity_sim = self.f_sim
                affinity_dis = self.f_dis
            else:
                _found = False
                for f in self.f_sim:
                    if affinity == f.__name__:
                        affinity_sim.append(f)
                        _found = True
                        break
                for f in self.f_dis:
                    if affinity == f.__name__ and not _found:
                        affinity_dis.append(f)
                        _found = True
                        break
                if not _found:
                    raise Exception('affinity must be at least one of the following string: ' + str(sim.get_all_functions_name() + dis.get_all_functions_name()))
        elif isinstance(affinity, list):
            if len(affinity) == 0:
                raise Exception('affinity must be at least one of the following string: ' + str(sim.get_all_functions_name() + dis.get_all_functions_name() + ['all']))
            
            if len(affinity) == 1 and affinity[0] == 'all':
                affinity_sim = self.f_sim
                affinity_dis = self.f_dis

            # prevent double value
            _affinity_set = set(affinity)
            affinity = list(_affinity_set)

            for i in self.f_sim:
                if i.__name__ in affinity:
                    affinity_sim.append(i)
            for i in self.f_dis:
                if i.__name__ in affinity:
                    affinity_dis.append(i)

            func_len = len(affinity_sim) + len(affinity_dis)
            if 'yuleq' in affinity:
                func_len -= 1

            if func_len < len(affinity):
                raise Exception('you have affinity that is not in the list of similarity or distance')

        # check if linkage is valid
        if isinstance(linkage, str):
            if linkage == 'all':
                linkage = self.linkage
            elif linkage in self.linkage:
                linkage = [linkage]
            else:
                raise Exception('linkage must be one of the following: ' + str(self.linkage))
        elif isinstance(linkage, list):
            if len(linkage) == 0:
                raise Exception('linkage must be one of the following: ' + str(self.linkage))
            for i in linkage:
                if i not in self.linkage:
                    raise Exception('linkage must be one of the following: ' + str(self.linkage))
        
        # check if use_sampling
        if use_sampling == 'random':
            if 'seed' in kwargs:
                df_sample = util.random_sampling_data(df, sample_rate, seed)    
            else: 
                df_sample = util.random_sampling_data(df, sample_rate)
            df = df_sample
        elif use_sampling == 'stratified':
            if 'seed' in kwargs:
                df_sample = util.stratified_sampling_data(df, sample_rate, seed)    
            else: 
                df_sample = util.stratified_sampling_data(df, sample_rate)
            df = df_sample
        elif use_sampling == 'none' or use_sampling == 'None':
            pass
        else:
            raise Exception('Value must either "none", "stratified", or "random"')

        # prepare affinity function
        def eq(x, y) -> None:
            pass

        # separate dataframe last column to true_values
        true_values = df.iloc[:, -1] 

        # initate place holder for every linkage
        linkage_list = []
        # loop for all linkage
        for l in tqdm(linkage):
            for equation in tqdm(affinity_sim, leave=False):
                # update eq function
                def eq(x, y):
                    # print(x, y)
                    cm = confusion_matrix(x, y).ravel()
                    # print(cm)
                    if len(cm) < 4:
                        cm = np.append(cm, np.zeros(4 - len(cm)))
                    
                    eq_value = equation(cm[3], cm[2], cm[1], cm[0], sum(cm))
                    if eq_value == None or not np.isfinite(eq_value):
                        eq_value = 0

                    eq_value = util.convertEquation(eq_value, type=2)

                    return eq_value

                # scipy
                res = fclusterdata(df.iloc[:, :-1].values, 1.0, metric=eq, method=l)

                score = homogeneity_completeness_v_measure(true_values, res)
                adj_rand_score = adjusted_rand_score(true_values, res)

                linkage_list.append([l, 'sim ' + equation.__name__, score[0], score[1], score[2], adj_rand_score])

            for equation in tqdm(affinity_dis, leave=False):
                # update eq function
                def eq(x, y):
                    # print(x, y)
                    cm = confusion_matrix(x, y).ravel()
                    # print(cm)
                    if len(cm) < 4:
                        cm = np.append(cm, np.zeros(4 - len(cm)))
                    
                    eq_value = equation(cm[3], cm[2], cm[1], cm[0], sum(cm))
                    if eq_value == None or not np.isfinite(eq_value):
                        eq_value = 0

                    return eq_value

                # scipy
                res = fclusterdata(df.iloc[:, :-1].values, 1.0, metric=eq, method=l)

                score = homogeneity_completeness_v_measure(true_values, res)
                adj_rand_score = adjusted_rand_score(true_values, res)

                linkage_list.append([l, 'dis ' + equation.__name__, score[0], score[1], score[2], adj_rand_score])

        # convert to dataframe
        linkage_df = pd.DataFrame(linkage_list, columns=['linkage', 'equation', 'homogeneity', 'completeness', 'v_measure', 'adjusted_rand_index'])
        # sort by v_measure
        sort_by_column = 'adjusted_rand_index'
        if 'sort_by' in kwargs:
            sort_by_column = kwargs['sort_by']
        self.linkage_df = linkage_df.sort_values(by=sort_by_column, ascending=False, ignore_index=True)

        # show result if show_result is True
        if self.show_result:
            print(self.linkage_df.head(self.result_count))

    # function to print out the result of linkage
    def get_result(self, csv: bool = False) -> pd.DataFrame:
        """
        Return the result of best similarity equation that match with the best linkage
        
        Returns
        -------
        result_df : pandas.DataFrame
            Dataframe that contains the result of best similarity equation that match with the best linkage
        """
        if self.linkage_df.empty:
            raise Exception('Fit your data first with \'fit\' method')
        if csv:
            self.linkage_df.to_csv('result.csv', index=False)

        return self.linkage_df  


import setuptools

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="measurenary",
    version="1.0.0",
    author="Muhammad Dwiki Ramdhani",
    author_email="dwikiramdhani53@gmail.com",
    description="A package to calculate and find the best similarity between binary data features",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://measurenary.readthedocs.io/en/latest/",
    packages=setuptools.find_packages(include=['measurenary']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],

    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.21.6',
        'pandas>=1.3.5',
        'scikit_learn',
        'setuptools>=42',
        'tqdm'
    ],
    test_suite='tests'
)
